#
#   Pyrex Parser
#

from string import join, replace
from types import ListType, TupleType
import Nodes
import ExprNodes

def unquote(s):
  #print "unquoting:", s ###
  q = s[:3]
  if q == '"""' or q == "'''":
    #print "...triple" ###
    return replace(s[3:-3], '\n', r'\n')
  q = s[:1]
  s = s[1:-1]
  if q == '"':
    #print "...double" ###
    return s
  #print "...single" ###
  i = 0
  n = len(s)
  while i < n:
    c = s[i]
    if c == '"':
      s = s[:i] + '\\' + s[i:]
      n = n + 1
      i = i + 1
    elif c == '\\':
      i = i + 1
    i = i + 1
  return s

def p_ident(s, message = "Expected an identifier"):
  if s.sy == 'IDENT':
    name = s.systring
    s.next()
    return name
  else:
    s.error(message)

def p_ident_list(s):
  names = []
  while s.sy == 'IDENT':
    names.append(s.systring)
    s.next()
    if s.sy <> ',':
      break
    s.next()
  return names

#------------------------------------------
#
#   Expressions
#
#------------------------------------------

def p_binop_expr(s, ops, p_sub_expr):
  #print "p_binop_expr:", ops, p_sub_expr ###
  n1 = p_sub_expr(s)
  #print "p_binop_expr(%s):" % p_sub_expr, s.sy ###
  while s.sy in ops:
    op = s.sy
    pos = s.position()
    s.next()
    n2 = p_sub_expr(s)
    n1 = ExprNodes.binop_node(pos, op, n1, n2)
  return n1

#test: and_test ('or' and_test)* | lambdef

def p_simple_expr(s):
  #return p_binop_expr(s, ('or',), p_and_test)
  return p_rassoc_binop_expr(s, ('or',), p_and_test)

def p_rassoc_binop_expr(s, ops, p_subexpr):
  n1 = p_subexpr(s)
  if s.sy in ops:
    pos = s.position()
    op = s.sy
    s.next()
    n2 = p_rassoc_binop_expr(s, ops, p_subexpr)
    n1 = ExprNodes.binop_node(pos, op, n1, n2)
  return n1

#and_test: not_test ('and' not_test)*

def p_and_test(s):
  #return p_binop_expr(s, ('and',), p_not_test)
  return p_rassoc_binop_expr(s, ('and',), p_not_test)

#not_test: 'not' not_test | comparison

def p_not_test(s):
  if s.sy == 'not':
    pos = s.position()
    s.next()
    return ExprNodes.NotNode(pos, operand = p_not_test(s))
  else:
    return p_comparison(s)

#comparison: expr (comp_op expr)*
#comp_op: '<'|'>'|'=='|'>='|'<='|'<>'|'!='|'in'|'not' 'in'|'is'|'is' 'not'

def p_comparison(s):
  n1 = p_bit_expr(s)
  if s.sy in comparison_ops:
    pos = s.position()
    op = p_cmp_op(s)
    n2 = p_bit_expr(s)
    n1 = ExprNodes.PrimaryCmpNode(pos, 
      operator = op, operand1 = n1, operand2 = n2)
    if s.sy in comparison_ops:
      n1.cascade = p_cascaded_cmp(s)
  return n1

def p_cascaded_cmp(s):
  pos = s.position()
  op = p_cmp_op(s)
  n2 = p_bit_expr(s)
  result = ExprNodes.CascadedCmpNode(pos, 
    operator = op, operand2 = n2)
  if s.sy in comparison_ops:
    result.cascade = p_cascaded_cmp(s)
  return result

def p_cmp_op(s):
  if s.sy == 'not':
    s.next()
    s.expect('in')
    op = 'not_in'
  elif s.sy == 'is':
    s.next()
    if s.sy == 'not':
      s.next()
      op = 'is_not'
    else:
      op = 'is'
  else:
    op = s.sy
    s.next()
  if op == '<>':
    op = '!='
  return op
  
comparison_ops = (
  '<', '>', '==', '>=', '<=', '<>', '!=', 
  'in', 'is', 'not'
)

#expr: xor_expr ('|' xor_expr)*

def p_bit_expr(s):
  return p_binop_expr(s, ('|',), p_xor_expr)

#xor_expr: and_expr ('^' and_expr)*

def p_xor_expr(s):
  return p_binop_expr(s, ('^',), p_and_expr)

#and_expr: shift_expr ('&' shift_expr)*

def p_and_expr(s):
  return p_binop_expr(s, ('&',), p_shift_expr)

#shift_expr: arith_expr (('<<'|'>>') arith_expr)*

def p_shift_expr(s):
  return p_binop_expr(s, ('<<', '>>'), p_arith_expr)

#arith_expr: term (('+'|'-') term)*

def p_arith_expr(s):
  return p_binop_expr(s, ('+', '-'), p_term)

#term: factor (('*'|'/'|'%') factor)*

def p_term(s):
  return p_binop_expr(s, ('*', '/', '%'), p_factor)

#factor: ('+'|'-'|'~'|typecast|sizeof) factor | power

def p_factor(s):
  if s.sy in ('+', '-', '~'):
    op = s.sy
    pos = s.position()
    s.next()
    return ExprNodes.unop_node(pos, op, p_factor(s))
  elif s.sy == "<":
    return p_typecast(s)
  elif s.sy == 'IDENT' and s.systring == "sizeof":
    return p_sizeof(s)
  else:
    return p_power(s)

def p_typecast(s):
  # s.sy == "<"
  pos = s.position()
  s.next()
  base_type = p_c_base_type(s)
  declarator = p_c_declarator(s, empty = 1)
  s.expect(">")
  operand = p_factor(s)
  return ExprNodes.TypecastNode(pos, 
    base_type = base_type, 
    declarator = declarator,
    operand = operand)

def p_sizeof(s):
  # s.sy == ident "sizeof"
  pos = s.position()
  s.next()
  s.expect('(')
  operand = p_simple_expr(s)
  s.expect(')')
  return ExprNodes.SizeofNode(pos, operand = operand)

#power: atom trailer* ('**' factor)*

def p_power(s):
  n1 = p_atom(s)
  while s.sy in ('(', '[', '.'):
    n1 = p_trailer(s, n1)
  if s.sy == '**':
    pos = s.position()
    s.next()
    n2 = p_factor(s)
    n1 = ExprNodes.binop_node(pos, '**', n1, n2)
  return n1

#trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME

def p_trailer(s, node1):
  pos = s.position()
  if s.sy == '(':
    return p_call(s, node1)
  elif s.sy == '[':
    s.next()
    index = p_subscript_list(s)
    s.expect(']')
    return ExprNodes.IndexNode(pos, 
      base = node1, index = index)
  else: # s.sy == '.'
    s.next()
    name = p_ident(s)
    return ExprNodes.AttributeNode(pos, 
      obj = node1, attribute = name)

# arglist:  argument (',' argument)* [',']
# argument: [test '='] test       # Really [keyword '='] test

def p_call(s, function):
  # s.sy == '('
  pos = s.position()
  s.next()
  positional_args = []
  keyword_args = []
  star_arg = None
  starstar_arg = None
  while s.sy not in ('*', '**', ')'):
    arg = p_simple_expr(s)
    if s.sy == '=':
      s.next()
      if not arg.is_name:
        s.error("Expected an identifier before '='",
          pos = arg.pos)
      keyword = ExprNodes.StringNode(arg.pos, 
        value = arg.name)
      arg = p_simple_expr(s)
      keyword_args.append((keyword, arg))
    else:
      if keyword_args:
        s.error("Non-keyword arg following keyword arg",
          pos = arg.pos)
      positional_args.append(arg)
    if s.sy <> ',':
      break
    s.next()
  if s.sy == '*':
    s.next()
    star_arg = p_simple_expr(s)
    if s.sy == ',':
      s.next()
  if s.sy == '**':
    s.next()
    starstar_arg = p_simple_expr(s)
    if s.sy == ',':
      s.next()
  s.expect(')')
  if not (keyword_args or star_arg or starstar_arg):
    return ExprNodes.SimpleCallNode(pos,
      function = function,
      args = positional_args)
  else:
    arg_tuple = None
    keyword_dict = None
    if positional_args or not star_arg:
      arg_tuple = ExprNodes.TupleNode(pos, 
        args = positional_args)
    if star_arg:
      if arg_tuple:
        arg_tuple = ExprNodes.binop_node(pos, 
          operator = '+', operand1 = arg_tuple,
          operand2 = star_arg)
      else:
        arg_tuple = star_arg
    if keyword_args:
      keyword_dict = ExprNodes.DictNode(pos,
        key_value_pairs = keyword_args)
    return ExprNodes.GeneralCallNode(pos, 
      function = function,
      positional_args = arg_tuple,
      keyword_args = keyword_dict,
      starstar_arg = starstar_arg)

#lambdef: 'lambda' [varargslist] ':' test

#subscriptlist: subscript (',' subscript)* [',']

def p_subscript_list(s):
  pos = s.position()
  expr = p_subscript(s)
  if s.sy <> ',':
    return expr
  else:
    items = [expr]
    while s.sy == ',':
      s.next()
      if s.sy == ']':
        break
      items.append(p_subscript(s))
    return ExprNodes.TupleNode(pos, args = items)

#subscript: '.' '.' '.' | test | [test] ':' [test] [sliceop]
#sliceop: ':' [test]

def p_subscript(s):
  #print "p_subscript" ###
  pos = s.position()
  if s.sy == '.':
    s.next()
    s.expect('.')
    s.expect('.')
    return ExprNodes.EllipsisNode(pos)
  else:
    start = None
    stop = None
    step = None
    if s.sy <> ':':
      #print "p_subscript: getting start" ###
      start = p_simple_expr(s)
      if s.sy <> ':':
        return start
    # s.sy == ':'
    #print "p_subscript: getting first colon" ###
    s.next()
    if s.sy <> ':':
      #print "p_subscript: getting stop" ###
      stop = p_simple_expr(s)
    if s.sy == ':':
      #print "p_subscript: getting second colon" ###
      s.next()
      if s.sy not in (',', ']'):
        #print "p_subscript: getting step" ###
        step = p_simple_expr(s)
    return ExprNodes.SliceNode(pos,
      start = start or ExprNodes.NoneNode(pos), 
      stop = stop or ExprNodes.NoneNode(pos), 
      step = step or ExprNodes.NoneNode(pos))

#atom: '(' [testlist] ')' | '[' [listmaker] ']' | '{' [dictmaker] '}' | '`' testlist '`' | NAME | NUMBER | STRING+

def p_atom(s):
  pos = s.position()
  sy = s.sy
  if sy == '(':
    s.next()
    if s.sy == ')':
      result = ExprNodes.TupleNode(pos, args = [])
    else:
      result = p_expr(s)
    s.expect(')')
    return result
  elif sy == '[':
    return p_list_maker(s)
  elif sy == '{':
    return p_dict_maker(s)
  elif sy == '`':
    return p_backquote_expr(s)
  elif sy == 'INT':
    value = int(s.systring)
    s.next()
    return ExprNodes.IntNode(pos, value=value)
  elif sy == 'FLOAT':
    value = float(s.systring)
    s.next()
    return ExprNodes.FloatNode(pos, value=value)
  elif sy == 'STRING':
    #value = s.systring
    strings = [unquote(s.systring)]
    s.next()
    while s.sy == 'STRING':
      strings.append(unquote(s.systring))
      s.next()
    value = join(strings, "")
    return ExprNodes.StringNode(pos, value = value)
  elif sy == 'IDENT':
    name = s.systring
    s.next()
    return ExprNodes.NameNode(pos, name=name)
  elif sy == '&':
    s.next()
    arg = p_atom(s)
    return ExprNodes.AmpersandNode(pos, operand = arg)
  else:
    s.error("Expected an identifier or literal")

def p_list_maker(s):
  # s.sy == '['
  pos = s.position()
  s.next()
  exprs = p_simple_expr_list(s)
  s.expect(']')
  return ExprNodes.ListNode(pos, args = exprs)

#dictmaker: test ':' test (',' test ':' test)* [',']

def p_dict_maker(s):
  # s.sy == '{'
  pos = s.position()
  s.next()
  items = []
  while s.sy <> '}':
    key = p_simple_expr(s)
    s.expect(':')
    value = p_simple_expr(s)
    items.append((key, value))
    if s.sy <> ',':
      break
    s.next()
  s.expect('}')
  return ExprNodes.DictNode(pos, key_value_pairs = items)

def p_backquote_expr(s):
  # s.sy == '`'
  pos = s.position()
  s.next()
  arg = p_expr(s)
  s.expect('`')
  return ExprNodes.BackquoteNode(pos, arg = arg)

#testlist: test (',' test)* [',']

def p_simple_expr_list(s):
  exprs = []
  while s.sy not in expr_terminators:
    exprs.append(p_simple_expr(s))
    if s.sy <> ',':
      break
    s.next()
  return exprs

def p_expr(s):
  pos = s.position()
  expr = p_simple_expr(s)
  if s.sy == ',':
    s.next()
    exprs = [expr] + p_simple_expr_list(s)
    return ExprNodes.TupleNode(pos, args = exprs)
  else:
    return expr

expr_terminators = (')', ']', '}', ':', '=', 'EOL')

#-------------------------------------------------------
#
#   Statements
#
#-------------------------------------------------------

def p_global_statement(s):
  # assume s.sy == 'global'
  pos = s.position()
  s.next()
  names = p_ident_list(s)
  return Nodes.GlobalNode(pos, names = names)

def p_expression_or_assignment(s):
  e1 = p_expr(s)
  if s.sy == '=':
    pos = s.position()
    s.next()
    e2 = p_expr(s)
    if (e1.is_sequence_constructor and	
        e2.is_sequence_constructor):
      pairs = []
      flatten_multiple_assignment(s, e1, e2, pairs)
      return Nodes.MultipleAssignmentNode(pos,
        lhs_rhs_pairs = pairs)
    else:
      return Nodes.SingleAssignmentNode(pos,
        lhs = e1, rhs = e2)
  else:
    return Nodes.ExprStatNode(e1.pos, expr = e1)

def flatten_multiple_assignment(s, lhs, rhs, pairs):
  # lhs and rhs are both SequenceNodes
  if len(lhs.args) <> len(rhs.args):
    s.error("Unpacking sequence of wrong size (expected %d, got %d)"
      % (len(lhs.args), len(rhs.args)), pos = rhs.pos)
  for i in range(len(lhs.args)):
    lhs_arg = lhs.args[i]
    rhs_arg = rhs.args[i]
    if (lhs_arg.is_sequence_constructor and	
        rhs_arg.is_sequence_constructor):
      flatten_multiple_assignment(s, lhs_arg, rhs_arg, pairs)
    else:
      pairs.append((lhs_arg, rhs_arg))

def p_print_statement(s):
  # s.sy == 'print'
  pos = s.position()
  s.next()
  if s.sy == '>>':
    s.error("'print >>' not yet implemented")
  args = []
  ewc = 0
  if s.sy <> 'NEWLINE':
    args.append(p_simple_expr(s))
    while s.sy == ',':
      s.next()
      if s.sy == 'NEWLINE':
        ewc = 1
        break
      args.append(p_simple_expr(s))
  return Nodes.PrintStatNode(pos, 
    args = args, ends_with_comma = ewc)

def p_del_statement(s):
  # s.sy == 'del'
  pos = s.position()
  s.next()
  args = p_simple_expr_list(s)
  return Nodes.DelStatNode(pos, args = args)

def p_pass_statement(s):
  # s.sy == 'pass'
  pos = s.position()
  s.next()
  return Nodes.PassStatNode(pos)

def p_break_statement(s):
  # s.sy == 'break'
  pos = s.position()
  s.next()
  return Nodes.BreakStatNode(pos)

def p_continue_statement(s):
  # s.sy == 'continue'
  pos = s.position()
  s.next()
  return Nodes.ContinueStatNode(pos)

def p_return_statement(s):
  # s.sy == 'return'
  pos = s.position()
  s.next()
  if s.sy not in statement_terminators:
    value = p_expr(s)
  else:
    value = None
  return Nodes.ReturnStatNode(pos, value = value)

def p_raise_statement(s):
  # s.sy == 'raise'
  pos = s.position()
  s.next()
  exc_type = None
  exc_value = None
  exc_tb = None
  if s.sy not in statement_terminators:
    exc_type = p_simple_expr(s)
    if s.sy == ',':
      s.next()
      exc_value = p_simple_expr(s)
      if s.sy == ',':
        s.next()
        exc_tb = p_simple_expr(s)
  return Nodes.RaiseStatNode(pos, 
    exc_type = exc_type,
    exc_value = exc_value,
    exc_tb = exc_tb)

def p_import_statement(s):
  # s.sy == 'import'
  pos = s.position()
  s.next()
  items = [p_dotted_name(s, as_allowed = 1)]
  while s.sy == ',':
    s.next()
    items.append(p_dotted_name(s, as_allowed = 1))
  stats = []
  for pos, target_name, dotted_name, as_name in items:
    stat = Nodes.SingleAssignmentNode(pos,
      lhs = ExprNodes.NameNode(pos, 
        name = as_name or target_name),
      rhs = ExprNodes.ImportNode(pos, 
        module_name = ExprNodes.StringNode(pos,
          value = dotted_name),
        name_list = None))
    stats.append(stat)
  return Nodes.StatListNode(pos, stats = stats)

def p_from_import_statement(s):
  # s.sy == 'from'
  pos = s.position()
  s.next()
  (dotted_name_pos, _, dotted_name, _) = \
    p_dotted_name(s, as_allowed = 0)
  s.expect('import')
  if s.sy == '*':
    s.error("'import *' not supported")
  imported_names = [p_imported_name(s)]
  while s.sy == ',':
    s.next()
    imported_names.append(p_imported_name(s))
  imported_name_strings = []
  items = []
  for (name_pos, name, as_name) in imported_names:
    imported_name_strings.append(
      ExprNodes.StringNode(name_pos, value = name))
    items.append(
      (name,
       ExprNodes.NameNode(name_pos, 
        name = as_name or name)))
  import_list = ExprNodes.ListNode(
    imported_names[0][0], args = imported_name_strings)
  return Nodes.FromImportStatNode(pos,
    module = ExprNodes.ImportNode(dotted_name_pos,
      module_name = ExprNodes.StringNode(dotted_name_pos,
        value = dotted_name),
      name_list = import_list),
    items = items)

def p_imported_name(s):
  pos = s.position()
  name = p_ident(s)
  as_name = p_as_name(s)
  return (pos, name, as_name)

def p_dotted_name(s, as_allowed):
  pos = s.position()
  target_name = p_ident(s)
  as_name = None
  names = [target_name]
  while s.sy == '.':
    s.next()
    names.append(p_ident(s))
  if as_allowed:
    as_name = p_as_name(s)
  else:
    as_name = None
  return (pos, target_name, join(names, "."), as_name)

def p_as_name(s):
  if s.sy == 'IDENT' and s.systring == 'as':
    s.next()
    return p_ident(s)
  else:
    return None

def p_assert_statement(s):
  # s.sy == 'assert'
  pos = s.position()
  s.next()
  expr = p_expr(s)
  return Nodes.AssertStatNode(pos, expr = expr)

statement_terminators = (';', 'NEWLINE', 'EOF')

def p_simple_statement(s):
  if s.sy == 'global':
    node = p_global_statement(s)
  elif s.sy == 'print':
    node = p_print_statement(s)
  elif s.sy == 'del':
    node = p_del_statement(s)
  elif s.sy == 'pass':
    node = p_pass_statement(s)
  elif s.sy == 'break':
    node = p_break_statement(s)
  elif s.sy == 'continue':
    node = p_continue_statement(s)
  elif s.sy == 'return':
    node = p_return_statement(s)
  elif s.sy == 'raise':
    node = p_raise_statement(s)
  elif s.sy == 'import':
    node = p_import_statement(s)
  elif s.sy == 'from':
    node = p_from_import_statement(s)
  elif s.sy == 'assert':
    node = p_assert_statement(s)
#	elif s.sy == 'xxx':
#		node = p_xxx_statement(s)
  else:
    node = p_expression_or_assignment(s)
  if s.sy == 'NEWLINE':
    s.next()
  else:
    s.error("Syntax error in statement")
  return node

def p_if_statement(s):
  # s.sy == 'if'
  pos = s.position()
  s.next()
  if_clauses = [p_if_clause(s)]
  while s.sy == 'elif':
    s.next()
    if_clauses.append(p_if_clause(s))
  else_clause = p_else_clause(s)
  return Nodes.IfStatNode(pos,
    if_clauses = if_clauses, else_clause = else_clause)

def p_if_clause(s):
  pos = s.position()
  test = p_simple_expr(s)
  body = p_suite(s)
  return Nodes.IfClauseNode(pos,
    condition = test, body = body)

def p_else_clause(s):
  if s.sy == 'else':
    s.next()
    return p_suite(s)
  else:
    return None

def p_while_statement(s):
  # s.sy == 'while'
  pos = s.position()
  s.next()
  test = p_simple_expr(s)
  body = p_suite(s)
  else_clause = p_else_clause(s)
  return Nodes.WhileStatNode(pos, 
    condition = test, body = body, 
    else_clause = else_clause)

def p_for_statement(s):
  # s.sy == 'for'
  pos = s.position()
  s.next()
  target = p_for_target(s)
  s.expect('in')
  iterator = p_for_iterator(s)
  body = p_suite(s)
  else_clause = p_else_clause(s)
  return Nodes.ForStatNode(pos, 
    target = target,
    iterator = iterator,
    body = body,
    else_clause = else_clause)

def p_for_target(s):
  pos = s.position()
  expr = p_bit_expr(s)
  if s.sy == ',':
    s.next()
    exprs = [expr]
    while s.sy <> 'in':
      exprs.append(p_bit_expr(s))
      if s.sy <> ',':
        break
      s.next()
    return ExprNodes.TupleNode(pos, args = exprs)
  else:
    return expr

def p_for_iterator(s):
  pos = s.position()
  expr = p_expr(s)
  return ExprNodes.IteratorNode(pos, sequence = expr)

def p_try_statement(s):
  # s.sy == 'try'
  pos = s.position()
  s.next()
  body = p_suite(s)
  except_clauses = []
  if s.sy == 'except':
    while s.sy == 'except':
      except_clauses.append(p_except_clause(s))
    return Nodes.TryExceptStatNode(pos,
      body = body, except_clauses = except_clauses)
  elif s.sy == 'finally':
    s.next()
    finally_clause = p_suite(s)
    return Nodes.TryFinallyStatNode(pos,
      body = body, finally_clause = finally_clause)
  else:
    s.error("Expected 'except' or 'finally'")

def p_except_clause(s):
  # s.sy == 'except'
  pos = s.position()
  s.next()
  exc_type = None
  exc_value = None
  if s.sy <> ':':
    exc_type = p_simple_expr(s)
    if s.sy == ',':
      s.next()
      exc_value = p_simple_expr(s)
  body = p_suite(s)
  return Nodes.ExceptClauseNode(pos,
    pattern = exc_type, target = exc_value, body = body)

def p_statement(s, level):
  if s.sy == 'cdef':
    if not (level == 'module' or level == 'function'):
      s.error('cdef statement not allowed here')
    return p_cdef_statement(s, level)
  elif s.sy == 'def':
    if not (level == 'module' or level == 'class'):
      s.error('def statement not allowed here')
    return p_def_statement(s)
  elif s.sy == 'class':
    if level <> 'module':
      s.error("class definition not allowed here")
    return p_class_statement(s)
  elif s.sy == 'if':
    return p_if_statement(s)
  elif s.sy == 'while':
    return p_while_statement(s)
  elif s.sy == 'for':
    return p_for_statement(s)
  elif s.sy == 'try':
    return p_try_statement(s)
#	elif s.sy == 'xxx':
#		return p_xxx_statement(s)
  else:
    return p_simple_statement(s)

def p_statement_list(s, level):
  pos = s.position()
  stats = []
  while s.sy <> 'EOF' and s.sy <> 'DEDENT':
    stats.append(p_statement(s, level))
  return Nodes.StatListNode(pos, stats = stats)

def p_suite(s, level = 'other'):
  pos = s.position()
  s.expect(':')
  stmts = []
  if s.sy == 'NEWLINE':
    s.next()
    #if s.sy <> 'INDENT':
    #	s.error("Expected an increase in indentation level")
    #s.next()
    s.expect_indent()
    result = p_statement_list(s, level)
    s.expect_dedent()
    #if s.sy <> 'DEDENT':
    #	s.error("Expected a decrease in indentation level")
    #s.next()
  else:
    result = p_simple_statement(s)
  return result

def p_c_base_type(s):
  signed = 1
  pos = s.position()
  if (s.sy == 'IDENT' 
      and (s.systring in s.type_names 
        or s.systring in base_type_start_words)):
    if s.systring == 'unsigned':
      signed = 0
      s.next()
      if s.sy <> 'IDENT':
        s.error('Expected a type name')
    name = s.systring
    s.next()
  else:
    name = "object"
  return Nodes.CBaseTypeNode(pos, signed = signed,
    name = name)

base_type_start_words = (
  "char", "short", "int", "long", "float", "double",
  "void", "unsigned", "object"
)

def p_c_declarator(s, empty = 0):
  pos = s.position()
  #print "p_c_declarator: looking for declarator" ###
  if s.sy == '*':
    #print "p_c_declarator: got *" ###
    s.next()
    base = p_c_declarator(s, empty)
    result = Nodes.CPtrDeclaratorNode(pos, 
      base = base)
  elif s.sy == '**':
    s.next()
    base = p_c_declarator(s, empty)
    result = Nodes.CPtrDeclaratorNode(pos,
      base = Nodes.CPtrDeclaratorNode(pos,
        base = base))
  else:
    if not empty:
      #print "p_c_declarator: looking for name" ###
      name = p_ident(s, 'Identifier missing in declarator')
    else:
      name = ""
    result = Nodes.CNameDeclaratorNode(pos, name = name)
    while s.sy in ('[', '('):
      if s.sy == '[':
        s.next()
        if s.sy <> ']':
          dim = p_expr(s)
        else:
          dim = None
        s.expect(']')
        result = Nodes.CArrayDeclaratorNode(pos, 
          base = result, dimension = dim)
      else: # sy == '('
        s.next()
        args = p_c_arg_list(s)
        s.expect(')')
        result = Nodes.CFuncDeclaratorNode(pos, 
          base = result, args = args)
  return result

def p_c_arg_list(s):
  args = []
  if s.sy <> ')':
    args.append(p_c_arg_decl(s))
    while s.sy == ',':
      s.next()
      args.append(p_c_arg_decl(s))
  return args

def p_c_arg_decl(s):
  pos = s.position()
  base_type = p_c_base_type(s)
  declarator = p_c_declarator(s)
  return Nodes.CArgDeclNode(pos,
    base_type = base_type,
    declarator = declarator)

def p_cdef_statement(s, level):
  # s.sy == 'cdef'
  pos = s.position()
  s.next()
  extern_flag = 0
  if s.sy == 'IDENT' and s.systring == "extern":
    extern_flag = 1
    s.next()
  if s.sy == 'IDENT' and s.systring in struct_union_or_enum:
    if level <> 'module':
      s.error("C struct/union/enum definition not allowed at this level")
    if s.systring == "enum":
      return p_c_enum_definition(s, pos)
    else:
      return p_c_struct_or_union_definition(s, pos)
  else:
    return p_c_func_or_var_declaration(s, pos, extern_flag)

struct_union_or_enum = (
  "struct", "union", "enum"
)

def p_c_enum_definition(s, pos):
  # s.sy == ident 'enum'
  s.next()
  name = p_ident(s)
  s.type_names.append(name)
  items = None
  s.expect(':')
  s.expect('NEWLINE')
  s.expect_indent()
  items = []
  while s.sy <> 'DEDENT':
    p_c_enum_line(s, items)
  s.expect_dedent()
  return Nodes.CEnumDefNode(pos, name = name, items = items)

def p_c_enum_line(s, items):
  p_c_enum_item(s, items)
  while s.sy == ',':
    s.next()
    if s.sy == 'NEWLINE':
      break
    p_c_enum_item(s, items)
  s.expect('NEWLINE')

def p_c_enum_item(s, items):
  pos = s.position()
  name = p_ident(s)
  value = None
  if s.sy == '=':
    s.next()
    value = p_simple_expr(s)
  items.append(Nodes.CEnumDefItemNode(pos, 
    name = name, value = value))

def p_c_struct_or_union_definition(s, pos):
  # s.sy == ident 'struct' or 'union'
  kind = s.systring
  s.next()
  name = p_ident(s)
  s.type_names.append(name)
  attributes = None
  if s.sy == ':':
    s.next()
    s.expect('NEWLINE')
    s.expect_indent()
    attributes = []
    while s.sy <> 'DEDENT':
      attributes.append(
        p_c_func_or_var_declaration(s, pos,
          suite_allowed = 0))
    s.expect_dedent()
  else:
    s.expect('NEWLINE')
  return Nodes.CStructOrUnionDefNode(pos, 
    name = name, kind = kind, attributes = attributes)

def p_c_func_or_var_declaration(s, pos, 
    extern_flag = 0, suite_allowed = 1):
  base_type = p_c_base_type(s)
  declarator = p_c_declarator(s)
  if s.sy == ':':
    if not suite_allowed:
      s.error("Suite not allowed here")
    suite = p_suite(s, 'function')
    result = Nodes.CFuncDefNode(pos,
      extern_flag = extern_flag,
      base_type = base_type,
      declarator = declarator, 
      body = suite)
  else:
    declarators = [declarator]
    while s.sy == ',':
      s.next()
      declarator = p_c_declarator(s)
      declarators.append(declarator)
    s.expect('NEWLINE', 
      'Syntax error in C variable declaration')
    result = Nodes.CVarDefNode(pos, 
      extern_flag = extern_flag,
      base_type = base_type, 
      declarators = declarators)
  return result

def p_def_statement(s):
  # assume s.sy == 'def'
  pos = s.position()
  s.next()
  name = p_ident(s)
  args = []
  s.expect('(');
  #args = p_ident_list(s)
  args = p_c_arg_list(s)
  s.expect(')')
  body = p_suite(s, 'function')
  return Nodes.DefNode(pos, name=name, args=args, body=body)

def p_class_statement(s):
  # s.sy == 'class'
  pos = s.position()
  s.next()
  class_name = p_ident(s)
  if s.sy == '(':
    s.next()
    base_list = p_simple_expr_list(s)
    s.expect(')')
  else:
    base_list = []
  body = p_suite(s, 'class')
  return Nodes.ClassDefNode(pos,
    name = class_name,
    bases = ExprNodes.TupleNode(pos, args = base_list),
    body = body)

def p_module(s):
  s.type_names = []
  pos = s.position()
  body = p_statement_list(s, 'module')
  if s.sy <> 'EOF':
    s.error("Syntax error in statement [%s,%s]" % (
      repr(s.sy), repr(s.systring)))
  return Nodes.ModuleNode(pos, body = body)

#----------------------------------------------
#
#   Debugging
#
#----------------------------------------------

def print_parse_tree(f, node, level, key = None):	
  ind = "  " * level
  if node:
    f.write(ind)
    if key:
      f.write("%s: " % key)
    t = type(node)
    if t == TupleType:
      f.write("(%s @ %s\n" % (node[0], node[1]))
      for i in xrange(2, len(node)):
        print_parse_tree(f, node[i], level+1)
      f.write("%s)\n" % ind)
      return
    elif isinstance(node, Node):
      try:
        tag = node.tag
      except AttributeError:
        tag = node.__class__.__name__
      f.write("%s @ %s\n" % (tag, node.pos))
      for name, value in node.__dict__.items():
        if name <> 'tag' and name <> 'pos':
          print_parse_tree(f, value, level+1, name)
      return
    elif t == ListType:
      f.write("[\n")
      for i in xrange(len(node)):
        print_parse_tree(f, node[i], level+1)
      f.write("%s]\n" % ind)
      return
  f.write("%s%s\n" % (ind, node))

